/**********************************************************************
Camera calibration info:
https://docs.opencv.org/4.0.1/d9/d0c/group__calib3d.html

Extrinsic parameters/matrix converts world coordinates to ideal camera coordinates
( x y z )' = R ( X Y Z )' + t

Rodrigues( rvec, R, J ) generates R from an rvec and the J = dRi/drvecj

xx = x/z
yy = y/z

Intrinsic parameters/matrix converts ideal camera coordinates to actual camera coordinates
k1, k2, k3, k4, k5, k6 are radial distortion coefficients
k1 < 0 barrel distortion
k1 > 0 pincushion distortion
p1, p2 are tangential distortion coefficients
s1, s2, s3, s4 are thin prism distortion coefficients

rr = xx*xx + yy*yy
kk = (1 + k1*rr + k2*rr*rr + k3*rr*rr*rr)/(1 + k4*rr + k5*rr*rr + k6*rr*rr*rr)

xxx = xx*kk + 2p1*xx*yy + p2*( rr + 2*xx*xx ) + s1*rr + s2*rr*rr;
yyy = yy*kk + p1*( rr + 2*yy*yy ) + 2p2*xx*yy + s3*rr + s4*rr*rr;

Intrinsic parameters/matrix converts camera coordinates to ideal image coordiinates
u = fx*xxx + cx
v = fy*yyy + cy

/**********************************************************************/

#include <iostream>
#include <sstream>
#include <string>
#include <ctime>
#include <cstdio>

#include <opencv2/core.hpp>
#include <opencv2/core/utility.hpp>
#include <opencv2/imgproc.hpp>
#include <opencv2/calib3d.hpp>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/videoio.hpp>
#include <opencv2/highgui.hpp>

/**********************************************************************/

using namespace cv;
using namespace std;

/**********************************************************************/

int nrFrames;
float squareSize;
double totalAvgErr;
Mat cameraMatrix, distCoeffs, rvec_tvec, imagePtMat;
Size imageSize, boardSize;
vector<Point3f> boardPoints;
FileStorage fs_images;
vector<string> imageList;
size_t atImageList;

/**********************************************************************/

void readCameraParams( string filename )
{
  FileStorage fs( filename, FileStorage::READ );

  fs[ "nr_of_frames" ] >> nrFrames;
  fs[ "image_width" ] >> imageSize.width;
  fs[ "image_height" ] >> imageSize.height;
  fs[ "board_width" ] >> boardSize.width;
  fs[ "board_height" ] >> boardSize.height;
  fs[ "square_size" ] >> squareSize;
  fs[ "camera_matrix" ] >> cameraMatrix;
  fs[ "distortion_coefficients" ] >> distCoeffs;
  fs[ "extrinsic_parameters" ] >> rvec_tvec;
  fs[ "image_points" ] >> imagePtMat;
  fs[ "grid_points" ] >> boardPoints ;

  fs.release();
}

/**********************************************************************/

static bool readStringList( const string& filename, vector<string>& l )
{
  l.clear();
  FileStorage fs_local(filename, FileStorage::READ);
  if( !fs_local.isOpened() )
    return false;
  FileNode n = fs_local.getFirstTopLevelNode();
  if( n.type() != FileNode::SEQ )
    {
      fs_local.release();
      return false;
    }
  FileNodeIterator it = n.begin(), it_end = n.end();
  for( ; it != it_end; ++it )
    l.push_back((string)*it);
  fs_local.release();
  return true;
}

/**********************************************************************/
// returns xxx.empty() on end of list

Mat nextImage()
{
  static bool first_call = true;

  if ( first_call )
    {
      if ( !readStringList( "iphone-images.xml", imageList ) )
	{
	  cout << "Bad image list.\n";
	  exit( -1 );
	}
      first_call = false;
    }

  Mat result, result2;
  if( atImageList < imageList.size() )
    result = imread(imageList[atImageList++], IMREAD_COLOR);
  // specify fx and fy and let the function compute the destination image size.
  if ( !result.empty() )
    {
      resize( result, result2, Size(), 0.5, 0.5, INTER_AREA );	
      return result2;
    }
  return result;
}

/**********************************************************************/

int main(int argc, char* argv[])
{

  readCameraParams( "iphone-data.xml" );
  cout << "nr_of_frames: " << nrFrames << endl;
  cout << "image_width: " << imageSize.width << endl;
  cout << "image_height: " << imageSize.height << endl;
  cout << "board_width: " << boardSize.width << endl;
  cout << "board_height: " << boardSize.height << endl;
  cout << "square_size: " << squareSize << endl;
  cout << "camera_matrix:\n" << cameraMatrix << endl;
  cout << "distortion_coefficients:\n" << distCoeffs << endl;
  cout << "extrinsic_parameters: " << rvec_tvec.size() << "\n" << rvec_tvec( Range( 0, 1 ), Range::all() )  << endl;
  cout << "image_points:" << imagePtMat.size() << "\n" << imagePtMat( Range( 0, 1 ), Range::all() ) << endl;
  cout << "grid_points:\n" << boardPoints.size() << "\n" <<  boardPoints[0] << endl;

  for( int k; k < 59; k++ )
    {
      Mat newImagePtMat = imagePtMat( Range( k, k+1 ), Range::all() );
      Mat newImagePtMat2 = Mat::zeros( 54, 2, CV_32F);
      for ( int i = 0; i < 54; ++i )
	{
	  for ( int j = 0; j < 2; ++j )
	    {
	      newImagePtMat2.at<float>(i,j) = newImagePtMat.at<float>(0,i*2 + j);
	    }
	}
      cout << "Checking rvec and tvec " << k << endl;
      // cout << newImagePtMat2.size() << "\n" << newImagePtMat2 << endl;

      // find extrinsic parameters
      Mat rvec, tvec;
      solvePnP( boardPoints, newImagePtMat2, cameraMatrix, distCoeffs, rvec, tvec, false );
      cout << rvec.size() << "\n" << rvec.t() << endl;
      cout << rvec_tvec( Range( k, k+1 ), Range( 0, 3 ) ) << endl;
      cout << tvec.size() << "\n" << tvec.t() << endl;
      cout << rvec_tvec( Range( k, k+1 ), Range( 3, 6 ) ) << "\n" << endl;
    }
  
  for( int k = 0; k < 59; k++ )
    {
      Mat image = nextImage();
      vector<Point2f> imagePoints;
      projectPoints( boardPoints,
		     rvec_tvec( Range( k, k+1 ), Range( 0, 3 ) ),
		     rvec_tvec( Range( k, k+1 ), Range( 3, 6 ) ),
		     cameraMatrix,
		     distCoeffs,
		     imagePoints );
      cout << "Point " << k << endl;
      cout << Mat(imagePoints).t() << endl;
      cout << imagePtMat( Range( k, k+1 ), Range::all() ) << "\n" << endl;
      cout << Mat(imagePoints).t() - imagePtMat( Range( k, k+1 ), Range::all() ) << "\n" << endl;
      Mat newImagePtMat = imagePtMat( Range( k, k+1 ), Range::all() );
      Mat newImagePtMat2 = Mat::zeros( 54, 2, CV_32F);
      for ( int i = 0; i < 54; ++i )
	{
	  circle( image, Point( newImagePtMat.at<float>(0,i*2), newImagePtMat.at<float>(0,i*2 + 1) ), 6, Scalar( 0, 0, 255 ), 2, LINE_AA  );
	}
      imshow( "Image", image );
      cout << "Press return to continue.\n";
      waitKey( 0 );
    }
}

/**********************************************************************/
